//
// RayDialer RX V1.1
//
// ASkr 11/2010, 05/2012
// www.askrprojects.net
//
// Notice:
//   If you don't own a C30 compiler license, use the provided binary.
//   The free version does not include the required optimization level!
//


#include "p33FJ12GP201.h"

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <math.h>



// CHANGES:
//
//  V1.1, 5/2012
//    - added magnitude output format (instead of power)
//		- added minimal I2C interface:
//      - only one shot DTMF acquisition supported
//      - no response to commands (OK/FAIL)
//      - only magnitude output supported
//      - upper four bits of status byte can not be used
//        (NXT I2C firmware bug in conjunction with dsPIC's
//         pull-low-after-ACK strategy)
//      - replaced all \r\n by \n and changed UART send routine;
//        (only ASCII is sent; there was no single "\r")
//
//  V1.0
//		- implemented sweep display (LCD)
//    - memory full; 5 bytes left; mission accomplished...
//
//  V0.6
//    - new hotkey "?":
//      Send help message out to serial...
//    - flash memory almost full; 153 bytes left ;-)
//
//  V0.5
//    - new hotkey "c":
//      Channel selection for sweep.
//      After "c" was pressed, hit 1-4 to select a channel for sweep mode.
//      ESC aborts.
//    - new hotkey "f":
//      Select sweep frequency range.
//      After "f" was pressed, hit 1-8 to select a tone (+-32 Hz analysis), "a" for all
//      or ESC to abort.
//    - changed (serial output) channel names to 1-4 (instead of 0-3)
//
//  V0.4
//    - initial version
//



// Hardware Setup dsPIC33FJ12GP201
// ===============================
//
//                        +-------\/-------+
//               pull H ->| MCLR       Vdd |<- VDD
//            channel 1 ->| AN0        Vss |<- GND
//            channel 2 ->| AN1       RP15 |-> TX
//            channel 3 ->| AN2       RP14 |<- RX
//            channel 0 ->| AN3    Vddcore |<- 10u
//            DEBUG PIN <-| RA2        Vss |<- GND
//              LCD CS  <-| RA3            |<> (res I2C, SDA)
//              LCD SCK <-| RP4            |<- (res I2C, SCL)
//              LCD CD  <-| RA4        RB7 |-> LCD SO
//                        +----------------+
//




//------------------------------------------------------------------------------------------------
// PIC STUFF
#define cOSCTUNE 12				// Can be determined by sweep analysis and known emitter frequency
													// As an alternative, the DTMF frequencies in gGOE_DTMF[] may be changed.


//------------------------------------------------------------------------------------------------
// DTMF STUFF
//   frequencies/Hz:
//     697  770  852  941 1209 1336 1477 1633
//   normalized coefficients :
//     c=2*cos(2*%pi*f/cGOE_SFRQ);
//   These values will be copied into "gGOE_cf[]" (faster calculations from RAM).
//
//
//   DTMF "numbers":
//
//            1209Hz 1336Hz 1477Hz 1633Hz
//     697Hz    1      2      3      A
//     770Hz    4      5      6      B
//     852Hz    7      8      9      C
//     941Hz    *      0      #      D
//
#define cGOE_TONES				8
#define cGOE_SWEEP_RANGE	48     // sweep fn-47..fn..fn+48 Hz 

// allows quick, manual tuning (RC osc!)
#define cGOE_F1   697
#define cGOE_F2   770
#define cGOE_F3   852
#define cGOE_F4   941
#define cGOE_F5  1209
#define cGOE_F6  1336
#define cGOE_F7  1477
#define cGOE_F8  1633
const unsigned int gGOE_DTMF[cGOE_TONES]={cGOE_F1,cGOE_F2,cGOE_F3,cGOE_F4,cGOE_F5,cGOE_F6,cGOE_F7,cGOE_F8};
const unsigned int gGOE_SWEEP[cGOE_TONES+1][2]=
	{{cGOE_F1-cGOE_SWEEP_RANGE+1,cGOE_F8+cGOE_SWEEP_RANGE},	// complete
	 {cGOE_F1-cGOE_SWEEP_RANGE+1,cGOE_F1+cGOE_SWEEP_RANGE}, // tone 1
	 {cGOE_F2-cGOE_SWEEP_RANGE+1,cGOE_F2+cGOE_SWEEP_RANGE}, // ...
	 {cGOE_F3-cGOE_SWEEP_RANGE+1,cGOE_F3+cGOE_SWEEP_RANGE},
	 {cGOE_F4-cGOE_SWEEP_RANGE+1,cGOE_F4+cGOE_SWEEP_RANGE},
	 {cGOE_F5-cGOE_SWEEP_RANGE+1,cGOE_F5+cGOE_SWEEP_RANGE},
	 {cGOE_F6-cGOE_SWEEP_RANGE+1,cGOE_F6+cGOE_SWEEP_RANGE},
	 {cGOE_F7-cGOE_SWEEP_RANGE+1,cGOE_F7+cGOE_SWEEP_RANGE},
	 {cGOE_F8-cGOE_SWEEP_RANGE+1,cGOE_F8+cGOE_SWEEP_RANGE}};// tone 8

#define cGOE_MODE_IDLE   0
#define cGOE_MODE_NORM   1		// standard operation; 4 channel, real time, DTMF tones
#define cGOE_MODE_SWEEP  2		// debug sweep; 1 channel, variable frequency
#define cGOE_MODE_NORM1  3		// one time: standard operation; 4 channel, real time, DTMF tones
#define cGOE_MODE_SWEEP1 4		// one time: debug sweep; 1 channel, variable frequency
#define cGOE_MODE_CHSLCT 5		// sweep channel selection mode; 1-4 selects channel, ESC aborts
#define cGOE_MODE_FSLCT  6		// sweep frequency selection mode; 1-8 selects tone, "a" for all, ESC aborts

#define cGOE_NORESET     0    // just a define...
#define cGOE_RESET       1    // just a define...


//------------------------------------------------------------------------------------------------
// GOERTZEL MEMORY
// Default settings take ~200ms for one complete cycle.
#define cGOE_MAXCHAN 4			// number of max channels (do not change)
#define cGOE_SFRQ 5000    	// sample frequency (*** USED! MUST MATCH TIMER SETTINGS! ***)
#define cGOE_SAMS 750				// initial number of steps for Goertzel algorithm (*** USED ***)
#define cGOE_INTMUL 1024		// integer-precision-multiplier ;-)
#define cGOE_OUTPUT_POW 1		// ouput as power
#define cGOE_OUTPUT_MAG 2		// ouput as magnitude

unsigned int            gGOE_sams = cGOE_SAMS;							// number of steps per Goertzel 
unsigned char						gGOE_channels;											// number of channels for analysis (<= cGOE_MAXCHAN)
unsigned char						gGOE_tones;													// number of tones for analysis (<= cGOE_TONES)
volatile unsigned int		gGOE_steps[cGOE_MAXCHAN];						// current amount of performed Goertzel steps
long										gGOE_cf[cGOE_TONES];								// normalized coefficients (see above)
unsigned int						gGOE_xt[cGOE_MAXCHAN];							// last sample (notice of change: unsigned int!)
long 										gGOE_s1[cGOE_MAXCHAN][cGOE_TONES];	// coefficient 1
long 										gGOE_s2[cGOE_MAXCHAN][cGOE_TONES];	// coefficient 2
volatile long 					gGOE_re[cGOE_MAXCHAN][cGOE_TONES];	// result of calculation
unsigned char           gGOE_outform = cGOE_OUTPUT_POW;			// output format

volatile unsigned int		gGOE_mode = cGOE_MODE_NORM;					// moved from main()

//------------------------------------------------------------------------------------------------
// SERIAL SWEEP OUTPUT
unsigned char						gSER_sweep1 = 1;										// if set, only one channel will be output to serial interface
unsigned char						gSER_sweep1ch = 0;									// channel for sweep mode
unsigned char						gSER_sweeprange = 0;								// sweep range (0=complete, 1=tone1, 2=tone2 ...)


//------------------------------------------------------------------------------------------------
// LCD
#define cLCD_COMPILE		// undefine for debug (saves memory)
#define cLCD_LIN 0
#define cLCD_LOG 1
unsigned long						gLCD_maxint = 0;										// scale value for horizontal intesity bar
unsigned char						gLCD_linlog = cLCD_LOG;							// default is logarithmic y axis
#ifdef cLCD_COMPILE
long										gLCD_sweepmem[cGOE_SWEEP_RANGE*2];	// LCD sweep display memory
#endif

//------------------------------------------------------------------------------------------------
// PINS
#define LCD_CD	LATAbits.LATA4
#define LCD_CS	LATAbits.LATA3
#define LCD_SO	LATBbits.LATB7
#define LCD_SCK	LATBbits.LATB4

#define DBGPIN	LATAbits.LATA2


//------------------------------------------------------------------------------------------------
// I2C
#define I2C_ADDRESS 0b1001000
#define I2C_ADDMASK 0b0000011
#define I2C_NONE    0 // I2C (data) write in progress
#define I2C_WRITE   1 // I2C (data) write in progress
#define I2C_READ    2 // I2C (data) read in progress
volatile unsigned char	gI2C_sysstat = 0;						// future upgrade
volatile unsigned char	gI2C_commaddr = 0;
volatile unsigned char  gI2C_lastmode = I2C_NONE;
volatile unsigned char	gI2C_buffer[2*cGOE_TONES];	// I2C comm buffer
volatile unsigned char	gI2C_bufferindex;						// current, transmitted value
volatile unsigned char  gI2C_cmd = 0;



//------------------------------------------------------------------------------------------------
// (some) prototypes (not all ;-)
// Decided to keep all this in a single module.
// ~1000 lines of code are not that much...
void goeReset(unsigned char channel);
void goeStep(unsigned char channel, unsigned char tone);



//------------------------------------------------------------------------------------------------
// MESSAGE TO BRAIN
#if cGOE_TONES != 8
#error cGOE_TONES must not be changed
#endif
#if cGOE_SFRQ != 5000
#error Changing cGOE_SFRQ requires changing timer settings too!
#endif
#if cGOE_INTMUL != 1024
#error You better keep your hands off cGOE_INTMUL...
#endif



//****************************************************************************
//*** picInit()
//***
//*** Initializes the CPU and HW
//****************************************************************************
void picInit()
{
	//---------------------------
	// oscillator
	//   NOTE: REMEMBER TO SET UART SPEED, IF ANYTHING CHANGES!
	__builtin_write_OSCCONH( OSCCON & 0b11111000); // turn off PLL for following changes
	__builtin_write_OSCCONL( OSCCON | 0b00000001); // update changes
	OSCTUNbits.TUN = cOSCTUNE;		// manually tuned to 8.00MHz
	CLKDIVbits.FRCDIV = 0;				// FRCDIV -> default; in/1 => 8.0MHz
	CLKDIVbits.PLLPRE = 0;				// PLLPRE -> default; in/2 => 4.0MHz
	PLLFBDbits.PLLDIV = 38;				// 4.0MHz * (38+2) = 160MHz (40MIPS)
//	PLLFBDbits.PLLDIV = 18;				// 4.0MHz * (18+2) = 80.0 MHz (20MIPS)
	CLKDIVbits.PLLPOST = 0b00;		// in/2 => 40.0 MHz
	__builtin_write_OSCCONH( OSCCON | 0b00000001); // turn on PLL
	__builtin_write_OSCCONL( OSCCON | 0b00000001); // update changes


	//---------------------------
	// pins and ports
	TRISA = 0x0003;								// RA0-1: A/D
	TRISB = 0x0003;								// RB0-1: A/D
	TRISBbits.TRISB14 = 1;				// RX; error in manual: RX pin will not automatically be claimed by UART!

	LCD_CS = 1;
	LCD_CD = 0;


	//---------------------------
	// PIN SELECT
	__builtin_write_OSCCONL( OSCCON & 0b10111111); // unlock peripheral pin select
	RPOR7bits.RP15R = 0b00011;		// map U1TX to RP15
	RPINR18bits.U1RXR= 0b01110;		// map U1RX to RP14
	RPOR3bits.RP7R = 0b00111;			// map SDO1 to RP7
	RPOR2bits.RP4R = 0b01000; 		// map SCK1OUT to RP4
	__builtin_write_OSCCONL( OSCCON | 0b01000000); // lock peripheral pin select


	//---------------------------
	// UART, 115200, 8N1, idle high
	U1STA =0b0010000000000000;		// error in dsPIC docs (UTXINV)
	U1MODE=0b1000100000001000;		// well... ;-) 
	U1BRG=86;											// 115200Baud@40MIPS settings
//	U1BRG=42;										// 115200Baud@20MIPS settings
	U1STAbits.UTXEN = 1;					// enable UART and claim pin
	
	
	//---------------------------
	// SPI
	// now coded manually (no programmable pin left for CS)
/*	
	SPI1CON1bits.CKE = 0;					// data changes from idle->active
	SPI1CON1bits.CKP = 1;					// idle state is high
	SPI1CON1bits.MSTEN = 1;				// master mode
	SPI1CON1bits.SPRE = 0b000;		// 8:1  2nd prescale
	SPI1CON1bits.PPRE = 0b00;			// 64:1 1st prescale
	SPI1STATbits.SPIEN = 1;				// enable SPI
	IPC2bits.SPI1IP = 0x07;	
	IFS0bits.SPI1IF = 0;
	IEC0bits.SPI1IE = 1;					// enable SPI interrupt
*/


	//---------------------------
	// interrupts
	INTCON1 = 0x8000;							// nested ints disabled


	//---------------------------
	// A/D
	AD1CON1bits.ADON=0;
	AD1CSSL  =0;									// scan off
	AD1PCFGL =0b1111111111110000; // PCFG0-3 analog inputs
  AD1CON1bits.ADSIDL = 1;				// stop if device in idle mode
  AD1CON1bits.AD12B = 0;				// 4x10bit mode
  AD1CON1bits.FORM = 0;					// unsigned int output
  AD1CON1bits.SSRC = 0;					// clearing SAMP bit starts conversion
  AD1CON1bits.SIMSAM = 1;				// simultaneous sampling
  AD1CON1bits.ASAM = 1;					// auto-restart sampling after conversion
  AD1CON1bits.SAMP = 1;					// well, let them sample ;-)
  AD1CON2bits.VCFG = 0;					// ref is AVdd and AVss
  AD1CON2bits.CSCNA = 0;				// no input scan
  AD1CON2bits.CHPS = 0b10;			// convert ch0-3
  AD1CON2bits.SMPI = 0;					// interrupt for each sample/conv
  AD1CON2bits.BUFM = 0;					// fill buffer from the beginning
  AD1CON2bits.ALTS = 0;					// use only sample A
  AD1CON3bits.ADRC=0;						// use system clock
  AD1CON3bits.SAMC=1;						// number of sampling TADs (already sampling all the time between interrupts...)
  AD1CON3bits.ADCS=2;						// 2+1=3; 4*25ns=75ns TAD cycle

	AD1CHS0bits.CH0NB = 0;				// ch0 B -: Vref-
	AD1CHS0bits.CH0SB = 0b00011;	// ch0 B +: AN3
	AD1CHS0bits.CH0NA = 0;				// ch0 A -: Vref-
	AD1CHS0bits.CH0SA = 0b00011;	// ch0 A +: AN3
	
	AD1CHS123bits.CH123NB = 0b00;	// ch1-3 B -: Vref- (NOTE: datasheet says 0b01 does the same)
	AD1CHS123bits.CH123SB = 0;    // ch1   B  : ch1: AN0; ch2: AN1; ch3: AN2
	AD1CHS123bits.CH123NA = 0b00; // ch1-3 A -: Vref- (NOTE: datasheet says 0b01 does the same)
	AD1CHS123bits.CH123SA = 0;		// ch1-3 A  : ch1: AN0; ch2: AN1; ch3: AN2

	AD1CON1bits.ADON=1;


	//---------------------------
	// timer 1 + interrupts
	//   - 100ns clock
	//   - 200us interrupt
	T1CONbits.TON = 0;			// disable timer
	T1CONbits.TCS = 0;			// internal clock
	T1CONbits.TGATE = 0;		// disable gate mode
	T1CONbits.TCKPS = 0b00;	// select 1:1 prescaler (25ns/cycle)
	TMR1 = 0x00;
//	PR1 = 10000;					// period value to 10000*25ns=250us (4kHz) @40MIPS
//	PR1 = 8000;						// period value to 8000*25ns=200us (5kHz) @40MIPS
	PR1 = 7797;							// period value to 8000*25ns=200us (5kHz) @40MIPS
//	PR1 =24000;						// !!!DEBUG ONLY!!! allows 4 115200k bytes transfers for debugging
//	PR1 = 3900;						// period value to 4000*50ns=200us (5kHz) @20MIPS
	IPC0bits.T1IP = 0x07;		// priority=7 (highest)
	IFS0bits.T1IF = 0;			// clear interrupt flag
	IEC0bits.T1IE = 0;			// disable interrupt (yet)


	//---------------------------
	// I2C
	I2C1CON = 0b1000001000000000;
//	I2C1CON = 0x8000;								// enable I2C module
//	I2C1CON = 0x9040;							// enable I2C module with clock-stretching

	I2C1ADD = I2C_ADDRESS;					// I2C address
	I2C1MSK = I2C_ADDMASK;					// address mask
	IFS1bits.SI2C1IF = 0;						// clear interrupt flag
//	pI2C_buffer = &gI2C_buffer[0];	// set buffer pointer to buffer

	// TODO: (I2C) enable interrupt somewhere else...	
	IEC1bits.SI2C1IE = 1;						// enable I2C interrupt
	
}



//****************************************************************************
//*** I2C INT
//***
//***
//****************************************************************************

void __attribute__((interrupt,no_auto_psv)) _SI2C1Interrupt(void)
{
	unsigned char tmp;
	long ltmp;

	// D_A  ->  DATA  !ADDRESS
	// R_W  ->  READ  !WRITE	
	// S    ->  START
	// P    ->  STOP

	// I2C1CONbits.SCLREL:
	//   If STREN == 0 (clock strech "off"):
	//   => HW clear #1: At beginning of slave data byte transmission
	//   => HW clear #2: At end of every address byte reception.


	if( I2C1STATbits.D_A == 0 )
	{
		// address byte detected
		tmp = I2C1RCV; // read address and empty receive buffer

		if( I2C1STATbits.R_W == 1 )
		{
			// read command
			
			gI2C_commaddr = ( tmp & 0b111 ) >> 1;
			
			for( tmp=0; tmp<cGOE_TONES; tmp++ )
			{
				if( gGOE_outform != cGOE_OUTPUT_MAG )
					ltmp = sqrt( gGOE_re[gI2C_commaddr][tmp] );
				else
					ltmp = gGOE_re[gI2C_commaddr][tmp];
				
				gI2C_buffer[(tmp<<1)+0] = ( ltmp >> 8 ) & 0xff;
				gI2C_buffer[(tmp<<1)+1] = ( ltmp      ) & 0xff;
			}

			gI2C_lastmode = I2C_READ;			
			I2C1TRN = gI2C_sysstat | 0b11110000; // NXT firmware bug hack
			gI2C_bufferindex = 0;
			I2C1CONbits.SCLREL = 1;
			
		}
		else
		{
			// write command
			gI2C_lastmode = I2C_WRITE;			
		}
		
	}
	else
	{
		// this was a data byte
		
		if( gI2C_lastmode == I2C_READ )
		{
			if( gI2C_bufferindex > 2 * cGOE_TONES - 1 )
				gI2C_bufferindex = 2 * cGOE_TONES - 1;
			I2C1TRN = gI2C_buffer[gI2C_bufferindex++];
			I2C1CONbits.SCLREL = 1;
		}
		else
		{
			gI2C_cmd = I2C1RCV;
		}


	}
	
	IFS1bits.SI2C1IF = 0;	// clear interrupt flag

}	






//****************************************************************************
//*** TIMER 1 INT
//***
//***
//****************************************************************************
void __attribute__((__interrupt__, __shadow__,__auto_psv__)) _T1Interrupt(void)
{
	unsigned char i,j;
	
	// start AD conversion
	AD1CON1bits.DONE=0;
  AD1CON1bits.SAMP = 0;
  while(!AD1CON1bits.DONE)
  {;}

	for(i=0;i<gGOE_channels;i++)
	{
	// make channel 0 left, channel 3 right
		gGOE_xt[gGOE_channels-i-1]=(*(&ADCBUF0+i));
		
		for(j=0;j<gGOE_tones;j++)
			goeStep(i,j);
		gGOE_steps[i]++;
	}

	//Clear Timer1 interrupt flag
	IFS0bits.T1IF = 0;
}



//****************************************************************************
//*** SPI INT
//***
//*** unused
//****************************************************************************
void __attribute__((__interrupt__, __shadow__,__auto_psv__)) _SPI1Interrupt(void)
{
	// clear SPI1 interrupt flag
	IFS0bits.SPI1IF = 0;

}



//**************************************************************************************
//*** serCheck()
//***
//*** Checks if a byte was received.
//*** RETURNS:
//***  -1   : nothing
//*** 0..255: received byte
//**************************************************************************************
int serCheck()
{
	
	// TODO
	if(U1STAbits.OERR)
	{
		U1STAbits.OERR=0;
		return -2;
	}
	
	if(U1STAbits.URXDA)
		return(U1RXREG & 0xff);
	else
		return -1;
}


//**************************************************************************************
//*** serFlush()
//***
//*** Flushes the receive buffer
//**************************************************************************************
void serFlush()
{
	while(serCheck() >= 0)
	{;}
}



//**************************************************************************************
//*** serSendByte()
//***
//*** Send a byte over the serial interface.
//**************************************************************************************
void serSendByte(unsigned char data)
{
	while(U1STAbits.UTXBF)
	{;}
	U1TXREG=data;
}


//**************************************************************************************
//*** serSendStr()
//***
//*** Send a string over the serial interface.
//*** str -> 0 terminated string
//**************************************************************************************
void serSendStr(char *str)
{
	while( *str != NULL )
	{
		while( U1STAbits.UTXBF )
		{;}
		
		if( *str == '\n' )
		{
			U1TXREG = '\r';
			while( U1STAbits.UTXBF )
			{;}
		}
		
		U1TXREG = *str++;
	}
}


//**************************************************************************************
//*** serHelpMsg()
//***
//*** Send a help message out to serial.
//**************************************************************************************
void serHelpMsg()
{
	// had to reduce the help screen (needed more memory)
	serSendStr("\n\nUSAGE:\n\n");
	serSendStr("ESC stop\n");
	serSendStr(" d  DTMF\n");
	serSendStr(" D  DTMF single\n");
	serSendStr(" s  sweep\n");
	serSendStr(" S  sweep single\n\n");
	serSendStr(" a  sweep all\n");
	serSendStr(" 0  sweep one\n\n");
	serSendStr(" c  sweep channel:\n");
	serSendStr("   1..4 ch1..4\n");
	serSendStr("   ESC  abort\n\n");
	serSendStr(" f  sweep range:\n");
	serSendStr("   1..8 f1..8\n");
	serSendStr("   a    all\n");
	serSendStr("   ESC  abort\n\n");
	serSendStr(" r  reset max\n\n");
	serSendStr(" l  LCD lin\n");
	serSendStr(" L  LCD log\n\n");
	serSendStr(" 1   250 steps\n");
	serSendStr(" 2   500 steps\n");
	serSendStr("...\n");
	serSendStr(" 9  2500 steps\n\n");
	serSendStr(" o  output mag\n");
	serSendStr(" O  output pow\n\n");

}




//**************************************************************************************
//*** clcLLtoA()
//***
//*** This might overflow!
//**************************************************************************************
char *clcLLtoA(char *buffer, long long i)
{
	unsigned long long n;
	unsigned int negate = 0;
	int c = 24;

	if (i < 0)
	{
		negate=1;
		n = -i;
	}
	else
		if (i == 0)
		{
			buffer[0] = '0';
			buffer[1] = 0;
			return buffer;
		}
		else
		{
			n = i;
		}
	
	buffer[c--] = 0;
	do
	{
		buffer[c--] = (n % 10) + '0';
		n = n / 10;
	} while (n);
	
	if (negate)
	{
		buffer[c--] = '-';
	}

	return &buffer[c+1];
}


//**************************************************************************************
//*** goeCalc()
//***
//*** Calculates the Gortzel frequency coefficients.
//*** Return value is
//***   cGOE_INTMUL*2*cos(2.0*3.141592654*(float)<freq>/(float)cGOE_SFRQ);
//**************************************0************************************************
long goeCalc(unsigned int freq)
{
	return (float)cos( 2.0 * 3.141592654 * freq / cGOE_SFRQ ) * 2 * cGOE_INTMUL;
}



//**************************************************************************************
//*** goeInitDTMF()
//***
//*** Initializes the Goertzel coefficients for DTMF recognition.
//*** <reset>: 0/1; determines Goertzel taps reset
//**************************************************************************************
void goeInitDTMF(unsigned char reset)
{
	unsigned char i;
	
	// calculate DTMF coefficients
	for( i=0; i<cGOE_TONES; i++ )
		gGOE_cf[i] = goeCalc( gGOE_DTMF[i] );

	// number of "tones" to analyze
	gGOE_tones = cGOE_TONES;

	// number of simultaneous sampled channels
	gGOE_channels = cGOE_MAXCHAN;
	
	// reset Goertzel taps?
	if( reset )
	{
		for( i=0; i<cGOE_MAXCHAN; i++ )
			goeReset( i );
	}
}



//**************************************************************************************
//*** goeInitSweep()
//***
//*** Initializes the Goertzel coefficients for sweep operation.
//*** <tones>     : number of tones to analyze during one acquisistion (max of cGOE_TONES)
//*** <start_freq>: start frequency; all other n tones will operate on 'start_freq+(n*step)'
//*** <step>      : frequency steps (1-255; and '0', well ;-)
//*** <reset>: 0/1; determines Goertzel taps reset
//**************************************************************************************
void goeInitSweep(unsigned char tones, unsigned int start_freq,
                  unsigned char step, unsigned char reset)
{
	unsigned char i;
	
	// calculate frequency coefficients
	for( i=0; i<tones; i++)
		gGOE_cf[i] = goeCalc( start_freq + (i * step) );

	// number of "tones" to analyze
	gGOE_tones = tones;

	// number of simultaneous sampled channels
	gGOE_channels = cGOE_MAXCHAN;
	
	// reset Goertzel taps?
	if( reset )
	{
		for( i=0; i<cGOE_MAXCHAN; i++ )
			goeReset(i);
	}

}



//**************************************************************************************
//*** goeReset()
//***
//*** This resets the Goertzel tap <channel>
//*** WARNING: No error checks performed!
//**************************************************************************************
void goeReset(unsigned char channel)
{
	unsigned char i;
	
	for( i=0; i<cGOE_TONES; i++ )
	{
		gGOE_s1[channel][i] = 0.0;
		gGOE_s2[channel][i] = 0.0;
		gGOE_steps[channel] = 0;
	}
	gGOE_xt[channel] = 0.0;
}



//**************************************************************************************
//*** goeStart()
//***
//*** Starts the interrupt driven sampling machine.
//**************************************************************************************
void goeStart()
{
	TMR1 = 0x00;						// clear timer value
	IFS0bits.T1IF = 0;			// clear interrupt flag
	IEC0bits.T1IE = 1;			// enable interrupt
	T1CONbits.TON = 1;			// enable timer
}



//**************************************************************************************
//*** goeStop()
//***
//*** Stops the interrupt driven sampling machine.
//**************************************************************************************
void goeStop()
{
	IEC0bits.T1IE = 1;			// disable interrupt
	T1CONbits.TON = 0;			// stop timer
	IFS0bits.T1IF = 0;			// clear interrupt flag (just in case)
}



//**************************************************************************************
//*** goeStep()
//***
//*** Performs exactly "one step" of the Goertzel algorithm.
//*** channel -> A/D channel
//*** tone    -> the frequency to analyze
//**************************************************************************************
void goeStep(unsigned char channel, unsigned char tone)
{
	long s;

	// s = xt + ( cf * s1 ) - s2;
	s =      gGOE_xt[channel]
	   + ( ( gGOE_cf[tone] * gGOE_s1[channel][tone] ) / cGOE_INTMUL )
	   -     gGOE_s2[channel][tone];
	  
	// s2 = s1;
	gGOE_s2[channel][tone] = gGOE_s1[channel][tone];
	
	// s1 = s;
	gGOE_s1[channel][tone] = s;
}



//**************************************************************************************
//*** goeResult()
//***
//*** Performs exactly "one step" of the Goertzel algorithm.
//*** channel -> A/D channel
//*** tone    -> the frequency to analyze
//**************************************************************************************
void goeResult(unsigned char channel, unsigned char tone)
{
	// r = (s1**2 + s2**2 - s1*s2*cf) / (SFREQ * SDURATION);
	// r = (s1**2 + s2**2 - s1*s2*cf) / (NUMBER_OF_SAMPLES);

	gGOE_re[channel][tone] =
			(float)(	 (long long)gGOE_s1[channel][tone] * gGOE_s1[channel][tone]
						  +  (long long)gGOE_s2[channel][tone] * gGOE_s2[channel][tone]
		 				  -( (float)gGOE_s1[channel][tone] * gGOE_s2[channel][tone] * gGOE_cf[tone] ) / cGOE_INTMUL ) / gGOE_sams; 
		 				  
	// if magnitute output selected, calc square root
	if( gGOE_outform == cGOE_OUTPUT_MAG )
		gGOE_re[channel][tone] = sqrt( gGOE_re[channel][tone] );
		 				  
}



//**************************************************************************************
//*** goeAcquisitionDTMF()
//***
//*** Performs a complete DTMF Goertzel analysis.
//**************************************************************************************
void goeAcquisitionDTMF()
{
	unsigned int i,j;

	char tmp[33];
	
	serSendStr("\nSTART_DTMF\n");
	
	// 8 tones/frequencies at once, on all 4 channels
	goeInitDTMF(cGOE_RESET);
	goeStart();
		
	while( gGOE_steps[cGOE_MAXCHAN-1] < gGOE_sams )
	{;}
	
	goeStop();
	
	for( j=0; j<4; j++ )
		for( i=0; i<8; i++ )
			goeResult(j,i);
		
	
	// TODO: UART output needs a better place
	for( j=0; j<4; j++ )
	{
		serSendStr( clcLLtoA( (char *)&tmp, j+1 ) );
		for( i=0; i<8; i++ )
		{
			serSendStr(",");
			serSendStr( clcLLtoA( (char *)&tmp, gGOE_re[j][i] ) );
		}
		serSendStr("\n");
	}

	serSendStr("STOP_DTMF\n");

}



//**************************************************************************************
//*** goeAcquisitionSweep8()
//***
//*** Performs a complete "free" sweeped Goertzel analysis of 8 tones.
//*** First tone is <freq>, the following 7 tones are <freq + n*step>, n=1,2,3,4,5,6,7
//**************************************************************************************
void goeAcquisitionSweep8(unsigned int freq, unsigned char step)
{
	unsigned int i,j;

	char tmp[33];

	// 8 tones/frequencies at once, from <freq> in <step> steps.
	goeInitSweep( 8, freq, step, cGOE_RESET);
	goeStart();
		
	while( gGOE_steps[0] < gGOE_sams )
	{;}

	goeStop();
	

	for( j=0; j<4; j++ )
		for( i=0; i<8; i++ )
			goeResult(j,i);


	// TODO: UART output needs a better place
	for( j = ( (!gSER_sweep1) ? 0 : gSER_sweep1ch); j < ( (!gSER_sweep1) ? 4 : gSER_sweep1ch+1); j++ )
	{
		for( i=0; i<8; i++ )
		{
			serSendStr( clcLLtoA( (char *)&tmp, j+1 ) );		
			serSendStr(",");
			serSendStr( clcLLtoA( (char *)&tmp, freq+i*step ) );		
			serSendStr(",");
			serSendStr( clcLLtoA( (char *)&tmp, gGOE_re[j][i] ) );		
			serSendStr("\n");
		}
	}
}



//**************************************************************************************
//*** goeAcquisitionSweep()
//***
//*** Performs a complete "free" sweeped Goertzel analysis between <sfreq> and ~<efreq>
//*** Resolution is 1 Hz. Number of analyses will always be n*8.
//**************************************************************************************
unsigned char goeAcquisitionSweep(unsigned int sfreq, unsigned int efreq)
{
	unsigned int i,j,n;
	unsigned char ret=0;

	serSendStr("\nSTART_SWEEP\n");
	
	n=0;
	for( i=sfreq; i<efreq; i+=8 )
	{
		// calc & serial output
		goeAcquisitionSweep8(i,1);

	// TODO: LCD code needs a better place and strategy
#ifdef cLCD_COMPILE

		// LCD hack
		if( (gSER_sweep1) && ( gSER_sweeprange > 0 ) )
		{
			for( j=0; j<8; j++ )
				gLCD_sweepmem[n+j] = gGOE_re[gSER_sweep1ch][j];
			n+=8;
		}
		
#endif		

		if( serCheck() == 27 )
		{
			ret = 1;
			break;	
		}	
	}

	serSendStr("STOP_SWEEP\n");

	return ret;
}


//**************************************************************************************
//*** spiSendByte
//***
//*** Manual SPI routine
//**************************************************************************************
void spiSendByte(unsigned char byte, unsigned char data)
{
	unsigned char i;
	
	if( data == 0 )
		LCD_CD = 0;
	else
		LCD_CD = 1;
		
	for( i=0; i<8; i++)
	{
		LCD_SCK = 0;
		if(byte & 0x80)
			LCD_SO = 1;
		else
			LCD_SO =0;
		LCD_SCK = 1;
		byte <<= 1;
	}
}


#ifdef cLCD_COMPILE
//**************************************************************************************
//*** lcdInit
//***
//*** initialize LCD
//**************************************************************************************
void lcdInit()
{
	const unsigned char init[] =
		{	0b01000000, 0xa0, 0xc8, 0xa4, 0xa6, 0xa2, 0x2f, 0x27, 0x81,
			0x10, 0xfa, 0x93, 0xaf, 0x00, 0x10 };
	
	unsigned char i;
	LCD_CS = 0;
	
/*
	new loop implementation saves ~30 bytes of memory
	
	spiSendByte(0b01000000 + 0,0);		// scroll line 0 (6 b)
	spiSendByte(0xa0,0);							// mirror seg 131..00 (1 b)
	spiSendByte(0xc8,0);							// mirror com 0xc8; no mirror 0xc0
	spiSendByte(0xa4,0);							// show SRAM content
	spiSendByte(0xa6,0);							// no inverse display (1 b)
	spiSendByte(0xa2,0);							// LCD bias 1/9
	spiSendByte(0x2f,0);							// booster, reg and follower on
	spiSendByte(0x27,0);							// res ratio max
	spiSendByte(0x81,0);							// contrast a
	spiSendByte(0x10,0);							// contrast b
	spiSendByte(0xfa,0);							// advanced
	spiSendByte(0x93,0);							// temp -0.11%/C; wrap column and page
	spiSendByte(0xaf,0);							// display on
	spiSendByte(0x00,0);							// SRAM col address L (4 b)
	spiSendByte(0x10,0);							// SRAM col address H (4 b)
*/	
	
	for( i=0;i<15;i++ )
		spiSendByte(init[i],0);
	
	LCD_CS = 1;
}
#endif // cLCD_COMPILE


#ifdef cLCD_COMPILE
//**************************************************************************************
//*** lcdFill
//***
//*** clear LCD
//**************************************************************************************
void lcdFill(unsigned char color)
{
	unsigned int i;
	
	LCD_CS = 0;
	
	if(color != 0)
		color=0xff;

	spiSendByte(0b10110000,0); // select first page

	for(i=0;i<132*8;i++)
		spiSendByte(color,1);

	LCD_CS = 1;
	
}
#endif // cLCD_COMPILE



#ifdef cLCD_COMPILE
//**************************************************************************************
//*** lcdDrawByte
//***
//*** Draws <byte> at position <x> on page <p>.
//**************************************************************************************
void lcdDrawByte(unsigned char byte, unsigned char x, unsigned char p)
{
	p &= 0x0f;
	if( x > 101 )
		x = 101;
	x += 30;

	LCD_CS = 0;
	
	spiSendByte( 0b10110000 + p, 0 );										// select page <p>
	spiSendByte( 0b00000000 + (x & 0x0f),0 );						// set colom low nibble address
	spiSendByte( 0b00010000 + ((x & 0xf0) >> 4 ), 0 );	// set colom low nibble address
	spiSendByte( byte, 1);															// draw byte

	LCD_CS = 1;
}
#endif // cLCD_COMPILE



#ifdef cLCD_COMPILE
//**************************************************************************************
//*** lcdDrawHBar
//***
//*** Draws an 8 pixel wide horizontal bar at page <p> (0..7)
//*** from left to right, width is <w>.
//**************************************************************************************
void lcdDrawHBar(unsigned char p, unsigned char w)
{
	unsigned char i, bars, rest;
	
	if( w > 102 )
		w = 102;
	p &= 0x07;
	
	LCD_CS = 0;
	
	for( i=0; i<w; i++ )
		lcdDrawByte( 0xff, i, p );
	for( i=w; i<102; i++ )
		lcdDrawByte( 0x00, i, p );

	LCD_CS = 1;
}
#endif // cLCD_COMPILE




#ifdef cLCD_COMPILE
//**************************************************************************************
//*** lcdDrawVBar
//***
//*** Draws a 2 pixel wide vertical bar at x position <x> from bottom
//*** to top (height <h>)
//**************************************************************************************
void lcdDrawVBar(unsigned char x, unsigned char h, unsigned char width)
{
	unsigned char i,bars,rest;
	
	if( h > 64 )
		h = 64;
	bars = h >> 3;		// divide by 8 (amount of full bytes from bottom)
	
	LCD_CS = 0;
	
	for( i=0; i<bars; i++ )
	{
		lcdDrawByte( 0xff, x ,7 - i );
		if( width == 2 )
			lcdDrawByte( 0xff, x + 1, 7 - i );
	}

	if( (rest = h % 8 ) )
	{
		rest = 8 - rest;
		lcdDrawByte( 0xff ^ (unsigned char)(pow(2,rest)-1), x  ,7 - bars );
		if(width == 2)
			lcdDrawByte( 0xff ^ (unsigned char)(pow(2,rest)-1), x + 1, 7 - bars );
	}

	LCD_CS = 1;
}
#endif // cLCD_COMPILE



#ifdef cLCD_COMPILE
//**************************************************************************************
//*** lcdDrawResultsDTMF
//***
//*** Draws 8 bars, representing intensity of all 8 channels, for each of the receivers.
//**************************************************************************************
void lcdDrawResultsDTMF()
{
	float fi;
	
	#define BARMAX 56
	
	unsigned char i, j, x;
	long m = 0;
	
	// calculate max value for normalization
	for( j=0; j<4; j++ )		// receiver number
		for( i=0; i<8; i++ )	// channel number
			if( gGOE_re[j][i] > m )
				m = gGOE_re[j][i];


	if( m > gLCD_maxint )
		gLCD_maxint = m;


	x = 2;
	for( j=0; j<4; j++ )
	{
		for( i=0; i<8; i++ )
		{
			if( gLCD_linlog == cLCD_LOG )
			{
				fi = BARMAX + 20 * log10( (float)gGOE_re[j][i] / (float)m );
				if( fi < 0.0 )
					fi = 0.0;
				lcdDrawVBar( x, (unsigned char)fi, 2 );
			}
			else
			{
				lcdDrawVBar( x, ( BARMAX * gGOE_re[j][i] ) / m, 2);
			}

			x += 2;
		}
		x += 11;
		
	}// END for all receivers

	// TODO: can overflow
	lcdDrawHBar( 0, (m*102) / gLCD_maxint );
	
}
#endif // cLCD_COMPILE


#ifdef cLCD_COMPILE
//**************************************************************************************
//*** lcdDrawResultsSweep
//***
//*** Draws 8 bars, representing intensity of all 8 channels, for each of the receivers.
//**************************************************************************************
void lcdDrawResultsSweep()
{
	float fi;
	
	#define BARMAX 56
	
	unsigned char i,j;
	long m = 0;
	
	// calculate max value for normalization
	for( i=0; i<2*cGOE_SWEEP_RANGE; i++)
		if( gLCD_sweepmem[i] > m )
			m = gLCD_sweepmem[i];

	if( m > gLCD_maxint )
		gLCD_maxint = m;

	for( i=0; i<2*cGOE_SWEEP_RANGE; i++ )
	{
			if( gLCD_linlog == cLCD_LOG )
			{
				fi=BARMAX + 20 * log10( (float)gLCD_sweepmem[i] / (float)m );
				if( fi < 0.0 )
					fi = 0.0;
				lcdDrawVBar( i, (unsigned char)fi, 1 );
			}
			else
				lcdDrawVBar( i, (BARMAX*gLCD_sweepmem[i]) / m, 1 );
	}

	lcdDrawHBar( 0, (m*102) / gLCD_maxint );
	
}
#endif // cLCD_COMPILE




//****************************************************************************
//*** main()
//***
//***
//****************************************************************************
int main(void)
{
	int ch;

	picInit();
	serFlush();
	
	
#ifdef cLCD_COMPILE
	
	lcdInit();
	lcdFill(0);
	
#endif // cLCD_COMPILE


	for(;;)
	{
		switch( gGOE_mode )
		{
			// -----------------------
			case cGOE_MODE_NORM:
			case cGOE_MODE_NORM1:
				{
					goeAcquisitionDTMF(); 
					
#ifdef cLCD_COMPILE						
					
					lcdFill(0);
					lcdDrawResultsDTMF();
					
#endif // cLCD_COMPILE

				} break;
			// -----------------------
			case cGOE_MODE_SWEEP:
			case cGOE_MODE_SWEEP1:
				{
					if( goeAcquisitionSweep( gGOE_SWEEP[gSER_sweeprange][0], gGOE_SWEEP[gSER_sweeprange][1]) )
						gGOE_mode = cGOE_MODE_IDLE;
						
#ifdef cLCD_COMPILE						
						
					lcdFill(0);
					if( (gSER_sweep1) && ( gSER_sweeprange > 0 ) )
						lcdDrawResultsSweep();
					else
						lcdFill(1);
						
#endif // cLCD_COMPILE

				}break;
			// -----------------------
		}


		if( ( gGOE_mode == cGOE_MODE_NORM1 ) || ( gGOE_mode == cGOE_MODE_SWEEP1 ) )
			gGOE_mode = cGOE_MODE_IDLE;

		if( gGOE_mode == cGOE_MODE_IDLE )
			serSendStr("IDLE\n");
			
		do
		{
			// Any incoming command ("key") from UART or I2C?
			if(  ( ( ch = serCheck() ) >= 0 ) || ( gI2C_cmd != 0 )  )
			{
				
				// was it an I2C command?
				if( gI2C_cmd != 0 )
				{
					ch = gI2C_cmd;
					gI2C_cmd = 0;
				}
				
				
				if( gGOE_mode == cGOE_MODE_CHSLCT )
				{
					switch( ch )
					{
						case '1': gSER_sweep1ch = 0; serSendStr("1\n");gGOE_mode = cGOE_MODE_IDLE;  break;
						case '2': gSER_sweep1ch = 1; serSendStr("2\n");gGOE_mode = cGOE_MODE_IDLE;  break;
						case '3': gSER_sweep1ch = 2; serSendStr("3\n");gGOE_mode = cGOE_MODE_IDLE;  break;
						case '4': gSER_sweep1ch = 3; serSendStr("4\n");gGOE_mode = cGOE_MODE_IDLE;  break;
						case 27 : gGOE_mode = cGOE_MODE_IDLE; serSendStr("ABORTED\n"); break;
					}
					continue;
				}
				
				if( gGOE_mode == cGOE_MODE_FSLCT )
				{
					switch( ch )
					{
						case '1': 
						case '2': 
						case '3': 
						case '4': 
						case '5': 
						case '6': 
						case '7': 
						case '8': gSER_sweeprange = ch-48; serSendByte(ch); serSendStr("\n"); gGOE_mode = cGOE_MODE_IDLE; break;
						case 'a': gSER_sweeprange = 0; serSendStr("ALL\n"); gGOE_mode = cGOE_MODE_IDLE;  break;
						case 27 : gGOE_mode = cGOE_MODE_IDLE; serSendStr("ABORTED\n"); break;
					}
					if( gSER_sweeprange > 8 )
						gSER_sweeprange = 0;
					
					continue;
				}
				
				
				switch( ch )
				{
					case '1': gGOE_sams =  250; serSendStr("SAMS=250\n");  break;
					case '2': gGOE_sams =  500; serSendStr("SAMS=500\n");  break;
					case '3': gGOE_sams =  750; serSendStr("SAMS=750\n");  break;
					case '4': gGOE_sams = 1000; serSendStr("SAMS=1000\n"); break;
					case '5': gGOE_sams = 1250; serSendStr("SAMS=1250\n"); break;
					case '6': gGOE_sams = 1500; serSendStr("SAMS=1500\n"); break;
					case '7': gGOE_sams = 1750; serSendStr("SAMS=1750\n"); break;
					case '8': gGOE_sams = 2000; serSendStr("SAMS=2000\n"); break;
					case '9': gGOE_sams = 2250; serSendStr("SAMS=2250\n"); break;
					case 'D': gGOE_mode = cGOE_MODE_NORM1; gLCD_maxint = 0; break;
					case 'S': gGOE_mode = cGOE_MODE_SWEEP1;gLCD_maxint = 0; break;
					case 'd': gGOE_mode = cGOE_MODE_NORM;  gLCD_maxint = 0; break;
					case 's': gGOE_mode = cGOE_MODE_SWEEP; gLCD_maxint = 0; break;
					case 'r': gLCD_maxint = 0; serSendStr("MAXINT=0\n"); break;
					case '0': gSER_sweep1 = 1; serSendStr("SWEEP=ONE\n"); break;
					case 'a': gSER_sweep1 = 0; serSendStr("SWEEP=ALL\n"); break;
					case 'l': gLCD_linlog = 0; serSendStr("LCD=LIN\n"); break;
					case 'L': gLCD_linlog = 1; serSendStr("LCD=LOG\n"); break;
					case 'c':	if ( gGOE_mode == cGOE_MODE_IDLE )
										{
											gGOE_mode = cGOE_MODE_CHSLCT;
											serSendStr("CHANNEL(1-4,ESC):");
										}	break;
					case 'f':	if ( gGOE_mode == cGOE_MODE_IDLE )
										{
											gGOE_mode = cGOE_MODE_FSLCT;
											serSendStr("SWEEPRANGE(1-8,a,ESC):");
										}	break;
					case 'o': gGOE_outform = cGOE_OUTPUT_MAG; gLCD_maxint = 0; serSendStr("OUT=MAG\n"); break;
					case 'O': gGOE_outform = cGOE_OUTPUT_POW; gLCD_maxint = 0; serSendStr("OUT=POW\n");break;
					case 'h':
					case '?':	if ( gGOE_mode == cGOE_MODE_IDLE )
										{
											serHelpMsg();
										}	break;
					case 27:  gGOE_mode = cGOE_MODE_IDLE; serSendStr("IDLE\n"); break;
				}// END switch
			}
		}while( ( gGOE_mode == cGOE_MODE_IDLE ) || ( gGOE_mode == cGOE_MODE_CHSLCT ) );
		
		
	}// END for

	
}


